@@ -1,27 +1,10 @@
-// Copyright, 2021, by Samuel G. D. Williams. <http://www.codeotaku.com>
-// 
-// Permission is hereby granted, free of charge, to any person obtaining a copy
-// of this software and associated documentation files (the "Software"), to deal
-// in the Software without restriction, including without limitation the rights
-// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
-// copies of the Software, and to permit persons to whom the Software is
-// furnished to do so, subject to the following conditions:
-// 
-// The above copyright notice and this permission notice shall be included in
-// all copies or substantial portions of the Software.
-// 
-// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
-// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
-// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
-// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
-// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
-// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
-// THE SOFTWARE.
+// Released under the MIT License.
+// Copyright, 2021-2025, by Samuel Williams.
 
 #include "uring.h"
 #include "selector.h"
-#include "list.h"
-#include "array.h"
+#include "../list.h"
+#include "../array.h"
 
 #include <liburing.h>
 #include <poll.h>
@@ -35,10 +18,9 @@
 enum {
 	DEBUG = 0,
 	DEBUG_COMPLETION = 0,
+	DEBUG_IO_READ = 1,
 };
 
-static VALUE IO_Event_Selector_URing = Qnil;
-
 enum {URING_ENTRIES = 64};
 
 #pragma mark - Data Type
@@ -140,11 +122,12 @@ size_t IO_Event_Selector_URing_Type_size(const void *_selector)
 	
 	return sizeof(struct IO_Event_Selector_URing)
 		+ IO_Event_Array_memory_size(&selector->completions)
+		+ IO_Event_List_memory_size(&selector->free_list)
 	;
 }
 
 static const rb_data_type_t IO_Event_Selector_URing_Type = {
-	.wrap_struct_name = "IO_Event::Backend::URing",
+	.wrap_struct_name = "IO::Event::Backend::URing",
 	.function = {
 		.dmark = IO_Event_Selector_URing_Type_mark,
 		.dcompact = IO_Event_Selector_URing_Type_compact,
@@ -152,7 +135,7 @@ static const rb_data_type_t IO_Event_Selector_URing_Type = {
 		.dsize = IO_Event_Selector_URing_Type_size,
 	},
 	.data = NULL,
-	.flags = RUBY_TYPED_FREE_IMMEDIATELY,
+	.flags = RUBY_TYPED_FREE_IMMEDIATELY | RUBY_TYPED_WB_PROTECTED,
 };
 
 inline static
@@ -228,7 +211,7 @@ VALUE IO_Event_Selector_URing_allocate(VALUE self) {
 	struct IO_Event_Selector_URing *selector = NULL;
 	VALUE instance = TypedData_Make_Struct(self, struct IO_Event_Selector_URing, &IO_Event_Selector_URing_Type, selector);
 	
-	IO_Event_Selector_initialize(&selector->backend, Qnil);
+	IO_Event_Selector_initialize(&selector->backend, self, Qnil);
 	selector->ring.ring_fd = -1;
 	
 	selector->pending = 0;
@@ -238,7 +221,10 @@ VALUE IO_Event_Selector_URing_allocate(VALUE self) {
 	
 	selector->completions.element_initialize = IO_Event_Selector_URing_Completion_initialize;
 	selector->completions.element_free = IO_Event_Selector_URing_Completion_free;
-	IO_Event_Array_allocate(&selector->completions, 1024, sizeof(struct IO_Event_Selector_URing_Completion));
+	int result = IO_Event_Array_initialize(&selector->completions, IO_EVENT_ARRAY_DEFAULT_COUNT, sizeof(struct IO_Event_Selector_URing_Completion));
+	if (result < 0) {
+		rb_sys_fail("IO_Event_Selector_URing_allocate:IO_Event_Array_initialize");
+	}
 	
 	return instance;
 }
@@ -249,7 +235,7 @@ VALUE IO_Event_Selector_URing_initialize(VALUE self, VALUE loop) {
 	struct IO_Event_Selector_URing *selector = NULL;
 	TypedData_Get_Struct(self, struct IO_Event_Selector_URing, &IO_Event_Selector_URing_Type, selector);
 	
-	IO_Event_Selector_initialize(&selector->backend, loop);
+	IO_Event_Selector_initialize(&selector->backend, self, loop);
 	int result = io_uring_queue_init(URING_ENTRIES, &selector->ring, 0);
 	
 	if (result < 0) {
@@ -291,7 +277,7 @@ VALUE IO_Event_Selector_URing_transfer(VALUE self)
 	struct IO_Event_Selector_URing *selector = NULL;
 	TypedData_Get_Struct(self, struct IO_Event_Selector_URing, &IO_Event_Selector_URing_Type, selector);
 	
-	return IO_Event_Selector_fiber_transfer(selector->backend.loop, 0, NULL);
+	return IO_Event_Selector_loop_yield(&selector->backend);
 }
 
 VALUE IO_Event_Selector_URing_resume(int argc, VALUE *argv, VALUE self)
@@ -315,7 +301,7 @@ VALUE IO_Event_Selector_URing_push(VALUE self, VALUE fiber)
 	struct IO_Event_Selector_URing *selector = NULL;
 	TypedData_Get_Struct(self, struct IO_Event_Selector_URing, &IO_Event_Selector_URing_Type, selector);
 	
-	IO_Event_Selector_queue_push(&selector->backend, fiber);
+	IO_Event_Selector_ready_push(&selector->backend, fiber);
 	
 	return Qnil;
 }
@@ -447,7 +433,7 @@ static
 VALUE process_wait_transfer(VALUE _arguments) {
 	struct process_wait_arguments *arguments = (struct process_wait_arguments *)_arguments;
 	
-	IO_Event_Selector_fiber_transfer(arguments->selector->backend.loop, 0, NULL);
+	IO_Event_Selector_loop_yield(&arguments->selector->backend);
 	
 	if (arguments->waiting->result) {
 		return IO_Event_Selector_process_status_wait(arguments->pid, arguments->flags);
@@ -484,6 +470,8 @@ VALUE IO_Event_Selector_URing_process_wait(VALUE self, VALUE fiber, VALUE _pid,
 		.fiber = fiber,
 	};
 	
+	RB_OBJ_WRITTEN(self, Qundef, fiber);
+	
 	struct IO_Event_Selector_URing_Completion *completion = IO_Event_Selector_URing_Completion_acquire(selector, &waiting);
 	
 	struct process_wait_arguments process_wait_arguments = {
@@ -560,7 +548,7 @@ VALUE io_wait_transfer(VALUE _arguments) {
 	struct io_wait_arguments *arguments = (struct io_wait_arguments *)_arguments;
 	struct IO_Event_Selector_URing *selector = arguments->selector;
 	
-	IO_Event_Selector_fiber_transfer(selector->backend.loop, 0, NULL);
+	IO_Event_Selector_loop_yield(&selector->backend);
 	
 	if (DEBUG) fprintf(stderr, "io_wait_transfer:waiting=%p, result=%d\n", (void*)arguments->waiting, arguments->waiting->result);
 	
@@ -587,6 +575,8 @@ VALUE IO_Event_Selector_URing_io_wait(VALUE self, VALUE fiber, VALUE io, VALUE e
 		.fiber = fiber,
 	};
 	
+	RB_OBJ_WRITTEN(self, Qundef, fiber);
+	
 	struct IO_Event_Selector_URing_Completion *completion = IO_Event_Selector_URing_Completion_acquire(selector, &waiting);
 	
 	struct io_uring_sqe *sqe = io_get_sqe(selector);
@@ -630,6 +620,7 @@ struct io_read_arguments {
 	struct IO_Event_Selector_URing *selector;
 	struct IO_Event_Selector_URing_Waiting *waiting;
 	int descriptor;
+	off_t offset;
 	char *buffer;
 	size_t length;
 };
@@ -643,11 +634,11 @@ io_read_submit(VALUE _arguments)
 	if (DEBUG) fprintf(stderr, "io_read_submit:io_uring_prep_read(waiting=%p, completion=%p, descriptor=%d, buffer=%p, length=%ld)\n", (void*)arguments->waiting, (void*)arguments->waiting->completion, arguments->descriptor, arguments->buffer, arguments->length);
 	
 	struct io_uring_sqe *sqe = io_get_sqe(selector);
-	io_uring_prep_read(sqe, arguments->descriptor, arguments->buffer, arguments->length, io_seekable(arguments->descriptor));
+	io_uring_prep_read(sqe, arguments->descriptor, arguments->buffer, arguments->length, arguments->offset);
 	io_uring_sqe_set_data(sqe, arguments->waiting->completion);
 	io_uring_submit_now(selector);
 	
-	IO_Event_Selector_fiber_transfer(selector->backend.loop, 0, NULL);
+	IO_Event_Selector_loop_yield(&selector->backend);
 	
 	return RB_INT2NUM(arguments->waiting->result);
 }
@@ -673,18 +664,21 @@ io_read_ensure(VALUE _arguments)
 }
 
 static int
-io_read(struct IO_Event_Selector_URing *selector, VALUE fiber, int descriptor, char *buffer, size_t length)
+io_read(struct IO_Event_Selector_URing *selector, VALUE fiber, int descriptor, char *buffer, size_t length, off_t offset)
 {
 	struct IO_Event_Selector_URing_Waiting waiting = {
 		.fiber = fiber,
 	};
 	
+	RB_OBJ_WRITTEN(selector->backend.self, Qundef, fiber);
+	
 	IO_Event_Selector_URing_Completion_acquire(selector, &waiting);
 	
 	struct io_read_arguments io_read_arguments = {
 		.selector = selector,
 		.waiting = &waiting,
 		.descriptor = descriptor,
+		.offset = offset,
 		.buffer = buffer,
 		.length = length
 	};
@@ -707,10 +701,11 @@ VALUE IO_Event_Selector_URing_io_read(VALUE self, VALUE fiber, VALUE io, VALUE b
 	size_t length = NUM2SIZET(_length);
 	size_t offset = NUM2SIZET(_offset);
 	size_t total = 0;
+	off_t from = io_seekable(descriptor);
 	
 	size_t maximum_size = size - offset;
 	while (maximum_size) {
-		int result = io_read(selector, fiber, descriptor, (char*)base+offset, maximum_size);
+		int result = io_read(selector, fiber, descriptor, (char*)base+offset, maximum_size, from);
 		
 		if (result > 0) {
 			total += result;
@@ -744,12 +739,52 @@ static VALUE IO_Event_Selector_URing_io_read_compatible(int argc, VALUE *argv, V
 	return IO_Event_Selector_URing_io_read(self, argv[0], argv[1], argv[2], argv[3], _offset);
 }
 
+VALUE IO_Event_Selector_URing_io_pread(VALUE self, VALUE fiber, VALUE io, VALUE buffer, VALUE _from, VALUE _length, VALUE _offset) {
+	struct IO_Event_Selector_URing *selector = NULL;
+	TypedData_Get_Struct(self, struct IO_Event_Selector_URing, &IO_Event_Selector_URing_Type, selector);
+	
+	int descriptor = IO_Event_Selector_io_descriptor(io);
+	
+	void *base;
+	size_t size;
+	rb_io_buffer_get_bytes_for_writing(buffer, &base, &size);
+	
+	size_t length = NUM2SIZET(_length);
+	size_t offset = NUM2SIZET(_offset);
+	size_t total = 0;
+	off_t from = NUM2OFFT(_from);
+	
+	size_t maximum_size = size - offset;
+	while (maximum_size) {
+		int result = io_read(selector, fiber, descriptor, (char*)base+offset, maximum_size, from);
+		
+		if (result > 0) {
+			total += result;
+			offset += result;
+			from += result;
+			if ((size_t)result >= length) break;
+			length -= result;
+		} else if (result == 0) {
+			break;
+		} else if (length > 0 && IO_Event_try_again(-result)) {
+			IO_Event_Selector_URing_io_wait(self, fiber, io, RB_INT2NUM(IO_EVENT_READABLE));
+		} else {
+			return rb_fiber_scheduler_io_result(-1, -result);
+		}
+		
+		maximum_size = size - offset;
+	}
+	
+	return rb_fiber_scheduler_io_result(total, 0);
+}
+
 #pragma mark - IO#write
 
 struct io_write_arguments {
 	struct IO_Event_Selector_URing *selector;
 	struct IO_Event_Selector_URing_Waiting *waiting;
 	int descriptor;
+	off_t offset;
 	char *buffer;
 	size_t length;
 };
@@ -763,11 +798,11 @@ io_write_submit(VALUE _argument)
 	if (DEBUG) fprintf(stderr, "io_write_submit:io_uring_prep_write(waiting=%p, completion=%p, descriptor=%d, buffer=%p, length=%ld)\n", (void*)arguments->waiting, (void*)arguments->waiting->completion, arguments->descriptor, arguments->buffer, arguments->length);
 	
 	struct io_uring_sqe *sqe = io_get_sqe(selector);
-	io_uring_prep_write(sqe, arguments->descriptor, arguments->buffer, arguments->length, io_seekable(arguments->descriptor));
+	io_uring_prep_write(sqe, arguments->descriptor, arguments->buffer, arguments->length, arguments->offset);
 	io_uring_sqe_set_data(sqe, arguments->waiting->completion);
 	io_uring_submit_pending(selector);
 	
-	IO_Event_Selector_fiber_transfer(selector->backend.loop, 0, NULL);
+	IO_Event_Selector_loop_yield(&selector->backend);
 	
 	return RB_INT2NUM(arguments->waiting->result);
 }
@@ -793,18 +828,21 @@ io_write_ensure(VALUE _argument)
 }
 
 static int
-io_write(struct IO_Event_Selector_URing *selector, VALUE fiber, int descriptor, char *buffer, size_t length)
+io_write(struct IO_Event_Selector_URing *selector, VALUE fiber, int descriptor, char *buffer, size_t length, off_t offset)
 {
 	struct IO_Event_Selector_URing_Waiting waiting = {
 		.fiber = fiber,
 	};
 	
+	RB_OBJ_WRITTEN(selector->backend.self, Qundef, fiber);
+	
 	IO_Event_Selector_URing_Completion_acquire(selector, &waiting);
 	
 	struct io_write_arguments arguments = {
 		.selector = selector,
 		.waiting = &waiting,
 		.descriptor = descriptor,
+		.offset = offset,
 		.buffer = buffer,
 		.length = length,
 	};
@@ -827,6 +865,7 @@ VALUE IO_Event_Selector_URing_io_write(VALUE self, VALUE fiber, VALUE io, VALUE
 	size_t length = NUM2SIZET(_length);
 	size_t offset = NUM2SIZET(_offset);
 	size_t total = 0;
+	off_t from = io_seekable(descriptor);
 	
 	if (length > size) {
 		rb_raise(rb_eRuntimeError, "Length exceeds size of buffer!");
@@ -834,7 +873,7 @@ VALUE IO_Event_Selector_URing_io_write(VALUE self, VALUE fiber, VALUE io, VALUE
 
 	size_t maximum_size = size - offset;
 	while (maximum_size) {
-		int result = io_write(selector, fiber, descriptor, (char*)base+offset, maximum_size);
+		int result = io_write(selector, fiber, descriptor, (char*)base+offset, maximum_size, from);
 		
 		if (result > 0) {
 			total += result;
@@ -868,6 +907,49 @@ static VALUE IO_Event_Selector_URing_io_write_compatible(int argc, VALUE *argv,
 	return IO_Event_Selector_URing_io_write(self, argv[0], argv[1], argv[2], argv[3], _offset);
 }
 
+VALUE IO_Event_Selector_URing_io_pwrite(VALUE self, VALUE fiber, VALUE io, VALUE buffer, VALUE _from, VALUE _length, VALUE _offset) {
+	struct IO_Event_Selector_URing *selector = NULL;
+	TypedData_Get_Struct(self, struct IO_Event_Selector_URing, &IO_Event_Selector_URing_Type, selector);
+	
+	int descriptor = IO_Event_Selector_io_descriptor(io);
+	
+	const void *base;
+	size_t size;
+	rb_io_buffer_get_bytes_for_reading(buffer, &base, &size);
+	
+	size_t length = NUM2SIZET(_length);
+	size_t offset = NUM2SIZET(_offset);
+	size_t total = 0;
+	off_t from = NUM2OFFT(_from);
+	
+	if (length > size) {
+		rb_raise(rb_eRuntimeError, "Length exceeds size of buffer!");
+	}
+
+	size_t maximum_size = size - offset;
+	while (maximum_size) {
+		int result = io_write(selector, fiber, descriptor, (char*)base+offset, maximum_size, from);
+		
+		if (result > 0) {
+			total += result;
+			offset += result;
+			from += result;
+			if ((size_t)result >= length) break;
+			length -= result;
+		} else if (result == 0) {
+			break;
+		} else if (length > 0 && IO_Event_try_again(-result)) {
+			IO_Event_Selector_URing_io_wait(self, fiber, io, RB_INT2NUM(IO_EVENT_WRITABLE));
+		} else {
+			return rb_fiber_scheduler_io_result(-1, -result);
+		}
+		
+		maximum_size = size - offset;
+	}
+	
+	return rb_fiber_scheduler_io_result(total, 0);
+}
+
 #endif
 
 #pragma mark - IO#close
@@ -901,24 +983,21 @@ struct __kernel_timespec * make_timeout(VALUE duration, struct __kernel_timespec
 		return NULL;
 	}
 	
-	if (FIXNUM_P(duration)) {
+	if (RB_INTEGER_TYPE_P(duration)) {
 		storage->tv_sec = NUM2TIMET(duration);
 		storage->tv_nsec = 0;
 		
 		return storage;
 	}
 	
-	else if (RB_FLOAT_TYPE_P(duration)) {
-		double value = RFLOAT_VALUE(duration);
-		time_t seconds = value;
-		
-		storage->tv_sec = seconds;
-		storage->tv_nsec = (value - seconds) * 1000000000L;
-		
-		return storage;
-	}
+	duration = rb_to_float(duration);
+	double value = RFLOAT_VALUE(duration);
+	time_t seconds = value;
 	
-	rb_raise(rb_eRuntimeError, "unable to convert timeout");
+	storage->tv_sec = seconds;
+	storage->tv_nsec = (value - seconds) * 1000000000L;
+	
+	return storage;
 }
 
 static
@@ -1007,7 +1086,7 @@ unsigned select_process_completions(struct IO_Event_Selector_URing *selector) {
 		if (waiting && waiting->fiber) {
 			assert(waiting->result != -ECANCELED);
 			
-			IO_Event_Selector_fiber_transfer(waiting->fiber, 0, NULL);
+			IO_Event_Selector_loop_resume(&selector->backend, waiting->fiber, 0, NULL);
 		}
 	}
 	
@@ -1026,7 +1105,7 @@ VALUE IO_Event_Selector_URing_select(VALUE self, VALUE duration) {
 	// Flush any pending events:
 	io_uring_submit_flush(selector);
 	
-	int ready = IO_Event_Selector_queue_flush(&selector->backend);
+	int ready = IO_Event_Selector_ready_flush(&selector->backend);
 	
 	int result = select_process_completions(selector);
 	
@@ -1046,14 +1125,14 @@ VALUE IO_Event_Selector_URing_select(VALUE self, VALUE duration) {
 		
 		if (!selector->backend.ready && !timeout_nonblocking(arguments.timeout)) {
 			struct timespec start_time;
-			IO_Event_Selector_current_time(&start_time);
+			IO_Event_Time_current(&start_time);
 			
 			// This is a blocking operation, we wait for events:
 			result = select_internal_without_gvl(&arguments);
 			
 			struct timespec end_time;
-			IO_Event_Selector_current_time(&end_time);
-			IO_Event_Selector_elapsed_time(&start_time, &end_time, &selector->idle_duration);
+			IO_Event_Time_current(&end_time);
+			IO_Event_Time_elapsed(&start_time, &end_time, &selector->idle_duration);
 			
 			// After waiting/flushing the SQ, check if there are any completions:
 			if (result > 0) {
@@ -1096,9 +1175,27 @@ VALUE IO_Event_Selector_URing_wakeup(VALUE self) {
 
 #pragma mark - Native Methods
 
+static int IO_Event_Selector_URing_supported_p(void) {
+	struct io_uring ring;
+	int result = io_uring_queue_init(32, &ring, 0);
+	
+	if (result < 0) {
+		rb_warn("io_uring_queue_init() was available at compile time but failed at run time: %s\n", strerror(-result));
+		
+		return 0;
+	}
+	
+	io_uring_queue_exit(&ring);
+	
+	return 1;
+}
+
 void Init_IO_Event_Selector_URing(VALUE IO_Event_Selector) {
-	IO_Event_Selector_URing = rb_define_class_under(IO_Event_Selector, "URing", rb_cObject);
-	rb_gc_register_mark_object(IO_Event_Selector_URing);
+	if (!IO_Event_Selector_URing_supported_p()) {
+		return;
+	}
+	
+	VALUE IO_Event_Selector_URing = rb_define_class_under(IO_Event_Selector, "URing", rb_cObject);
 	
 	rb_define_alloc_func(IO_Event_Selector_URing, IO_Event_Selector_URing_allocate);
 	rb_define_method(IO_Event_Selector_URing, "initialize", IO_Event_Selector_URing_initialize, 1);
@@ -1123,6 +1220,8 @@ void Init_IO_Event_Selector_URing(VALUE IO_Event_Selector) {
 #ifdef HAVE_RUBY_IO_BUFFER_H
 	rb_define_method(IO_Event_Selector_URing, "io_read", IO_Event_Selector_URing_io_read_compatible, -1);
 	rb_define_method(IO_Event_Selector_URing, "io_write", IO_Event_Selector_URing_io_write_compatible, -1);
+	rb_define_method(IO_Event_Selector_URing, "io_pread", IO_Event_Selector_URing_io_pread, 6);
+	rb_define_method(IO_Event_Selector_URing, "io_pwrite", IO_Event_Selector_URing_io_pwrite, 6);
 #endif
 	
 	rb_define_method(IO_Event_Selector_URing, "io_close", IO_Event_Selector_URing_io_close, 1);
